Attribute VB_Name = "modVariables"
'Variable Module
'
'This module contains the functions for dealing with variables, as well as the
'variable container, which is 'dVar'.

Option Explicit

Public Type typeVar
    Name As String
    Value As String
    Parent As String
    PID As Integer  'parent ID
    
    isArray As Boolean
    arrayValues() As String
    numElements As Integer
End Type

Public dVars() As typeVar, VarCount As Integer
Private VarErrorReason As String

Public Function AddVar(ByVal Name As String, ByVal Value As String, ByVal Parent As String, ByVal PID As Integer, doEval As Boolean) As Integer
    'adds a new variable to the var array
    Dim temp As String, varIndex As Integer
    Dim arrayMode As Boolean, numElements As Integer
    Dim i As Integer
    
    Name = Trim(Name)
    varIndex = -1
    
    If InStr(1, Name, "[") <> 0 Then
        parseArrayVar Name, numElements
        arrayMode = True
    End If
    
    If Not isValidVar(Name) Then _
        FlagError "Syntax Error", "Invalid variable name '" & Name & "'", iLocation: Exit Function
    
    If GetVarIndex(Name, False) <> -1 Then _
        FlagError "Run-Time Error", "Cannot create variable '" & Name & "'.  The variable already exists.", iLocation: Exit Function
    
    If VarCount <> 0 Then
        For i = 0 To VarCount
            If dVars(i).Name = "" And dVars(i).PID = -1 Then
                varIndex = i
            End If
        Next i
    End If
    If varIndex = -1 Then
        VarCount = VarCount + 1
        varIndex = VarCount
        ReDim Preserve dVars(VarCount)
    End If
    
    dVars(varIndex).Name = Name
    dVars(varIndex).Parent = Parent
    dVars(varIndex).PID = PID

    If debugMode Then _
        addDebug "Var +> " & Name & IIf(arrayMode, "[" & numElements & "]", "") & " -> " & Parent & " (pid: " & PID & ")"
    
    If arrayMode Then
        dVars(varIndex).numElements = numElements
        dVars(varIndex).isArray = True
        
        If numElements >= 0 Then
            ReDim dVars(varIndex).arrayValues(numElements)
        End If
    Else
        If doEval And Value <> "" Then
            dVars(varIndex).Value = Solve(Value)
        Else
            dVars(varIndex).Value = Value
        End If
    End If
    
    AddVar = varIndex
    
    If debugMode Then
        addDebug "    +> " & IIf(arrayMode, "(array, " & dVars(varIndex).numElements & ")", IIf(dVars(varIndex).Value <> "", dVars(varIndex).Value, "(no value)"))
        addDebug "    +> created"
    End If
        
End Function


Public Function GetVar(ByVal Name As String, Optional ByVal fromWhere As String) As String
    'gets the value of a variable
    Dim i As Integer
    Dim arrayMode As Boolean, elementNum As Integer
    
    Name = Trim(Name)
    
    If IsNumeric(Name) Then GetVar = Val(Name): Exit Function
    If Name = "" Then Exit Function
    
    If InStr(1, Name, "[") <> 0 Then
        parseArrayVar Name, elementNum
        arrayMode = True
    End If
    
    i = GetVarIndex(Name, False)
    If i = -1 Then
        Select Case Name
            Case "dpp.ip"
                GetVar = GetIPAddress
                AddVar "dpp.ip", GetIPAddress, "dpp.global", -1, False
            Case "dpp.host"
                GetVar = GetIPHostName
                AddVar "dpp.host", GetIPHostName, "dpp.global", -1, False
            Case "dpp.tick"
                GetVar = GetTickCount
            Case "dpp.time"
                GetVar = Time
            Case "dpp.vars"
                GetVar = VarCount
            Case Else
                If LCase(Name) = LCase(CurrentProc) And LCase(CurrentProc) <> "main" Then
                    'um, ok, this doesn't account for arrays, however _
                    'we should never get here *if the function is an array* _
                    'in the first place... it won't do any harm, _
                    'and it's late, and i don't feel like adding the check in.
                    'so screw it.
                    GetVar = ProcReturns(ReturnCount).Value
                    Exit Function
                Else
                    FlagError "Syntax Error", "Undefined variable '" & Name & "'" & vbCrLf & "Reason: " & VarErrorReason, iLocation: Exit Function
                End If
        End Select
        Exit Function
    End If
    
    If arrayMode Then
        checkArrayBounds dVars(i).numElements, elementNum, dVars(i).Name
        GetVar = dVars(i).arrayValues(elementNum)
    Else
        If dVars(i).isArray Then FlagError "Run-Time Error", "Variable type mismatch: " & vbCrLf & dVars(i).Name & ": dvar <-> darray", iLocation
        GetVar = dVars(i).Value
    End If
End Function

Public Sub SetVar(ByVal theVar As String, ByVal NewVal As String, Optional ByVal Parent As String)
    'sets the value of a variable
    Dim i As Integer
    Dim arrayMode As Boolean, elementNum As Integer
    
    theVar = Trim(theVar)
    
    If IsNumeric(theVar) Then FlagError "Run-Time Error", "Cannot modify variable '" & theVar & "'", iLocation: Exit Sub
    If Trim(theVar) = "" Then Exit Sub
    'theVar = Trim(theVar)
    
    If InStr(1, theVar, "[") <> 0 Then
        parseArrayVar theVar, elementNum
        arrayMode = True
    End If
    
    i = GetVarIndex(theVar, False)
    If i = -1 Then
        If theVar = CurrentProc And CurrentProc <> "main" Then
            'um, ok, this doesn't account for arrays, however _
            'we should never get here *if the function is an array* _
            'in the first place... it won't do any harm, _
            'and it's late, and i don't feel like adding the check in.
            'so screw it. (i copied this from above :p)
            ProcReturns(ReturnCount).Value = NewVal
            Exit Sub
        Else
            FlagError "Run-Time Error", "Undefined variable '" & theVar & "'" & vbCrLf & "Reason: " & VarErrorReason, iLocation: Exit Sub
        End If
    End If
    
    If Mid(theVar, 1, 4) = "dpp." Or LCase(theVar) = "true" Or LCase(theVar) = "false" Then
        FlagError "Run-Time Error", "Cannot modify variable '" & theVar & "'.  The variable is a system variable.", iLocation: Exit Sub
    End If
    
    If arrayMode Then
        checkArrayBounds dVars(i).numElements, elementNum, dVars(i).Name
        dVars(i).arrayValues(elementNum) = NewVal
    Else
        If dVars(i).isArray Then FlagError "Run-Time Error", "Variable type mismatch: " & vbCrLf & dVars(i).Name & ": dvar <-> darray", iLocation
        dVars(i).Value = NewVal
    End If
    If Parent <> "" Then dVars(i).Parent = Parent

End Sub

Public Function FindVar(ByVal theVar As String) As Boolean
    'determines if a variable exists
    Dim i As Integer, elementNum As Integer
    
    theVar = Trim(theVar)
    
    If theVar = "" Then Exit Function
   
    If InStr(1, theVar, "[") <> 0 Then _
        parseArrayVar theVar, elementNum
        
    VarErrorReason = ""
    For i = 1 To VarCount
        If LCase(dVars(i).Name) = LCase(theVar) Then
            'If LCase(dVars(i).Parent) = LCase(CurrentProc) Or Left(dVars(i).Parent, 4) = "dpp." Then
            If dVars(i).PID = CurrentPID Or dVars(i).PID = -1 Then
                FindVar = True
                VarErrorReason = ""
                Exit Function
            Else
                VarErrorReason = "Possible variable ownership mismatch " & vbCrLf & "(" & CurrentProc & " <-> " & dVars(i).Parent & ")"
            End If
        End If
    Next i
    
    If VarErrorReason = "" Then
        VarErrorReason = "Variable not found in stack"
    End If
    
    FindVar = False
End Function

Public Function GetVarIndex(ByVal theVar As String, displayError As Boolean, Optional elementNum As Integer) As Integer
    'determines if a variable exists
    Dim i As Integer
    
    theVar = Trim(theVar)
    VarErrorReason = ""
    elementNum = -1
    If theVar = "" Then GoTo finish
    
    If InStr(1, theVar, "[") <> 0 Then _
        parseArrayVar theVar, elementNum
    
    For i = 1 To VarCount
        If LCase(dVars(i).Name) = LCase(theVar) Then
            'If LCase(dVars(i).Parent) = LCase(CurrentProc) Or Left(dVars(i).Parent, 4) = "dpp." Then
            If dVars(i).PID = CurrentPID Or dVars(i).PID = -1 Then
                GetVarIndex = i
                VarErrorReason = ""
                Exit Function
            Else
                VarErrorReason = "Possible variable ownership mismatch " & vbCrLf & "(" & CurrentProc & " <-> " & dVars(i).Parent & ")"
            End If
        End If
    Next i
    
finish:
    If VarErrorReason = "" Then VarErrorReason = "Variable not found in stack"
    
    If displayError Then _
        FlagError "Run-Time Error", "Undefined variable '" & theVar & "'" & vbCrLf & "Reason: " & VarErrorReason, iLocation
    
    GetVarIndex = -1
End Function

Public Sub KillVar(ByVal varname As String)
    'erases a variable
    Dim i As Integer
    
    If Not FindVar(varname) Then Exit Sub
    For i = 1 To VarCount
        If LCase(dVars(i).Name) = LCase(varname) Then
            dVars(i).Name = ""
            dVars(i).Value = ""
            dVars(i).Parent = ""
            dVars(i).isArray = False
            dVars(i).numElements = 0
            ReDim dVars(i).arrayValues(0)
        End If
    Next i
End Sub

Public Sub ClearVars(ByVal ProcPID As Integer)
    'clears all vars from a certain function
    Dim i As Integer
    
    For i = 1 To VarCount
        If dVars(i).PID = ProcPID Then
            dVars(i).Name = ""
            dVars(i).Value = ""
            dVars(i).Parent = ""
            dVars(i).PID = -1
            dVars(i).isArray = False
            dVars(i).numElements = 0
            Erase dVars(i).arrayValues
        End If
    Next i
End Sub

Public Sub ClearAllVars()
    'clears all variables
    'For i = 1 To VarCount
    '    dVars(i).Name = ""
    '    dVars(i).Value = ""
    '    dVars(i).Parent = ""
    'Next i
    ReDim dVars(0)
    
    VarCount = 0
End Sub

Public Function isValidChar(ByVal Char As String) As Boolean
    'determines if a character is valid
    Dim charCode As Integer
    charCode = Asc(Mid(Char, 1, 1))
    
    If charCode >= 65 And charCode <= 90 Then GoTo validChar      'upper case
    If charCode >= 97 And charCode <= 122 Then GoTo validChar     'lower case
    If charCode >= 48 And charCode <= 57 Then GoTo validChar      'numbers
    If charCode = 95 Or charCode = 46 Then GoTo validChar        '_ & .
    If charCode = 91 Or charCode = 93 Then GoTo validChar        ' [ ]
    
    Exit Function
    
validChar:
    isValidChar = True
End Function

Public Function isValidVar(ByVal theVar As String) As Boolean
    'determines if a variable is in valid syntai
    Dim charCode As Integer, i As Integer
    
    If Trim(theVar) = "" Then GoTo invalidVar
    If IsNumeric(theVar) Then GoTo invalidVar
    If IsNumeric(Left(theVar, 1)) Then GoTo invalidVar
    If Len(theVar) > 128 Then GoTo invalidVar
    
    For i = 1 To Len(theVar)
        If Not isValidChar(Mid(theVar, i, 1)) Then Exit Function
    Next i
    
    isValidVar = True
    Exit Function
    
invalidVar:
End Function

Public Function parseArrayVar(Name As String, elementNum As Integer)
    Dim temp As String, n As Integer, m As Integer
    n = InStr(1, Name, "[")
    'm = InStr(n, Name, "]")
    m = FindEncloseEnd(n + 1, Name, "[")
    
    If m = 0 Then _
        FlagError "Syntax Error", "Expected ']'", iLocation
    
    temp = Mid(Name, n + 1, m - (n + 1))
    If temp <> "" Then
        Name = Trim(Mid(Name, 1, n - 1))
        If Val(temp) < 0 Then FlagError "Run-Time Error", "Array '" & Name & "': range has no values", iLocation
    End If
    elementNum = IIf(temp = "", -1, Val(Solve(temp)))
    Name = Trim(Mid(Name, 1, n - 1))
End Function

Public Function checkArrayBounds(numElements As Integer, elementIndex As Integer, arrName As String)
    If elementIndex > numElements Then GoTo flag
    If elementIndex < 0 Then GoTo flag
    
    Exit Function
flag:
    FlagError "Run-Time Error", "Subscript out of range.  " & vbCrLf & "Array: '" & arrName & "' Index: '" & elementIndex & "'", iLocation
End Function

Public Function makeIntoVar(theThing As String) As typeVar
    Dim tempInt As Integer
    Dim tempVar As typeVar, elementNum As Integer
    
    tempInt = GetVarIndex(theThing, False, elementNum)
    If tempInt <> -1 And elementNum = -1 Then
        tempVar = dVars(tempInt)
    Else
        tempVar.Value = Solve(theThing)
    End If
    
    makeIntoVar = tempVar
End Function
